#!/bin/bash

### The postprocessing part of the neural network. This is run at each generation, after all individual chunks are
### completed.

### 1. Loops the directories of each chunk and collects the cost value. Appends this to a copy of 00.csv and saves as a
###    new file, so that there will ultimately be a .csv file of parameters and their associated costs at every
###    generation
### 2. Performs a genetic algorithm (Biased Random Key Genetic Algorithm (http://mauricio.resende.info/doc/brkgaAPI.pdf)
###    the set of parameters that are generated are saved onto 00.csv, to be used in the next generation.


### 1. Loops the directories of each chunk and collects the cost value. Appends this to a copy of 00.csv and saves as a
###    new file, so that there will ultimately be a .csv file of parameters and their associated costs at every
###    generation

INPUT=/gpfs/scratch/bsc32/bsc32786/%EXPID%/EXP00/

module load python

function makecsv {

parameters="$(python - $1 << END
# $1 - path to directories of the chunks
# copies 00.csv to a new file (numbered according to generation number), makes a new column then loops through each
# chunk, reads cost.txt, adds each to the column

import numpy as np
import os
import pandas

dir='$1'
print(dir)
os.chdir(dir)

chunks=os.listdir(os.getcwd())
chunks=[i for i in chunks if os.path.isdir(i) if len(i)==2]

os.chdir(dir+'parameters')
parameters=np.genfromtxt('00.csv')
parameters=np.append(parameters,np.zeros((np.shape(parameters)[0],1)),axis=1)

for chunk in chunks:
    ind=int(chunk)-1
    os.chdir(dir+'/'+chunk)
    f=open('cost.txt','r')
    cost=float(f.read())
    f.close()
    print(cost)
    parameters[ind,-1]=cost
print(parameters)

os.chdir(dir+'/parameters')
gen=len(os.listdir(dir+'/parameters'))
print(gen)
gen="{0:0=2d}".format(gen)+'.csv'

np.savetxt(gen,parameters)

END
)"
}

makecsv $INPUT


### 2. Performs a genetic algorithm (Biased Random Key Genetic Algorithm (http://mauricio.resende.info/doc/brkgaAPI.pdf)
###    the set of parameters that are generated are saved onto 00.csv, to be used in the next generation.

SCRHOME=/gpfs/scratch/bsc32/bsc32786/
MODEL=%EXPID%/EXP00/
PARAMETERS=parameters/
cd $SCRHOME$MODEL$PARAMETERS

module load python

function GA {

GA="$(python - $1 $2 << END

import numpy as np
import random
import pandas
import os

def geneticalgorithm(input):

# Biased Random Key Genetic Algorithm (BRKGA)
# One generation crossover/mutation
# input - dataframe ((P+1)*M), extra column for values of the cost function
# output - dataframe (P*M) of new parameters

# P = number of parameters
# N = number of chunks

    #(N,P) = input.shape
    (N,P) = np.shape(input)
    P=P-1

    E=0.15 # proportion of elite
    M=0.1  # proportion of mutants
    rho=0.7 # probability of each crossover being assigned to the elite chunk

    NE=int(N*E)+1 # number of elite chunks (rounded up)
    NM=int(N*M)+1 # number of mutant chunks (rounded up)
    NC=N-(NE+NM) # number of crossover chunk

    # sorting the input by cost and selecting elite and non-elite chunks
    input=pandas.DataFrame(input)
    input=input.sort_values(input.shape[1]-1).values
    inputE=input[:NE,:]
    inputnE=input[NE:,:]

    # putting elite chunks into output
    output=input[:NE,:-1]

    # generating mutants and putting into output

    vars=pandas.read_csv('$2')

    for i in range(NM):
        r=np.zeros((1,P))
        for j in range(len(vars)):
            r[0,j]=random.random()*(vars['max'][j]-vars['min'][j])+vars['min'][j]
        output=np.append(output,r,axis=0)

    # crossover
    for i in range(NC):
        r=np.zeros((1,P))
        for j in range(P):
            en=inputE[np.random.randint(inputE.shape[0], size=1), :]
            nen=inputnE[np.random.randint(inputnE.shape[0], size=1), :]
            a=random.random()
            if a>rho:
                r[0,j]=nen[0,j]
            else:

                r[0,j]=en[0,j]
        output=np.append(output,r,axis=0)

    return output

os.chdir('$1')
gen=len(os.listdir('$1'))-1
gen="{0:0=2d}".format(gen)+'.csv'

A=np.genfromtxt(gen)

B=geneticalgorithm(A)
if os.path.exists("00.csv"):
    os.remove("00.csv")
np.savetxt('00.csv',B)

END
)"
}

GA $SCRHOME$MODEL$PARAMETERS $SCRHOME${MODEL}variables.csv